from ._base import BaseCapillarity

__all__ = [
    "TRUST",
]


class TRUST(BaseCapillarity):
    """TRUST capillary pressure.

    After Narasimhan et al. (1978).

    Parameters
    ----------
    p0 : scalar
        Capillary pressure strength (CP(1)).
    slr : scalar
        Irreducible liquid saturation (CP(2)).
    eta : scalar
        CP(3).
    pe : scalar
        Capillary entry pressure (CP(4)).
    pmax : scalar
        Maximum pressure (CP(5)).

    """

    _id = 3
    _name = "TRUST"

    def __init__(self, p0, slr, eta, pe, pmax):
        assert slr >= 0.0
        assert eta != 0.0
        self.parameters = [p0, slr, eta, pe, pmax]

    def _eval(self, sl, p0, slr, eta, pe, pmax):
        """TRUST capillary pressure."""
        if sl > slr:
            pcap = -pe - p0 * ((1.0 - sl) / (sl - slr)) ** (1.0 / eta)
        else:
            pcap = -abs(pmax)
        pcap = max(pcap, -abs(pmax))
        pcap *= (1.0 - sl) / 0.001 if sl > 0.999 else 1.0
        return pcap

    @property
    def parameters(self):
        """Return model parameters."""
        return [self._p0, self._slr, self._eta, self._pe, self._pmax]

    @parameters.setter
    def parameters(self, value):
        assert len(value) == 5
        self._p0 = value[0]
        self._slr = value[1]
        self._eta = value[2]
        self._pe = value[3]
        self._pmax = value[4]



from ._base import BaseRelativePermeability

__all__ = [
    "FattKlikoff",
]


class FattKlikoff(BaseRelativePermeability):
    """Fatt and Klikoff's function.

    After Fatt and Klikoff (1959).

    Parameters
    ----------
    slr : scalar
        Irreducible liquid saturation (RP(1)).

    """

    _id = 7
    _name = "Fatt-Klikoff"

    def __init__(self, slr):
        assert slr < 1.0
        self.parameters = [slr]

    def _eval(self, sl, slr):
        """Fatt and Klikoff's function."""
        Seff = (sl - slr) / (1.0 - slr) if sl > slr else 0.0
        kl = Seff ** 3
        kg = (1.0 - Seff) ** 3
        return kl, kg

    @property
    def parameters(self):
        """Return model parameters."""
        return [self._slr]

    @parameters.setter
    def parameters(self, value):
        assert len(value) == 1
        self._slr = value[0]



from ._base import BaseRelativePermeability

__all__ = [
    "Pickens",
]


class Pickens(BaseRelativePermeability):
    """Gas perfect mobile function.

    Parameters
    ----------
    x : scalar
        RP(1).

    """

    _id = 2
    _name = "Pickens"

    def __init__(self, x):
        self.parameters = [x]

    def _eval(self, sl, x):
        """Gas perfect mobile function."""
        return sl ** x, 1.0

    @property
    def parameters(self):
        """Return model parameters."""
        return [self._x]

    @parameters.setter
    def parameters(self, value):
        assert len(value) == 1
        self._x = value[0]



from ._base import BaseCapillarity

__all__ = [
    "vanGenuchten",
]


class vanGenuchten(BaseCapillarity):
    """van Genuchten function.

    After van Genuchten (1980).

    Parameters
    ----------
    m : scalar
        Related to pore size distribution (CP(1)).
    slr : scalar
        Irreducible liquid saturation (CP(2)).
    alpha : scalar
        Inverse of capillary pressure strength 1/P0 (CP(3)).
    pmax : scalar
        Maximum pressure (CP(4)).
    sls : scalar
        Maximum liquid saturation (CP(5)).

    """

    _id = 7
    _name = "van Genuchten"

    def __init__(self, m, slr, alpha, pmax, sls):
        assert pmax >= 0.0
        self.parameters = [m, slr, alpha, pmax, sls]

    def _eval(self, sl, m, slr, alpha, pmax, sls):
        """van Genuchten function."""
        if sl == 1.0:
            pcap = 0.0
        else:
            if sl > slr:
                Seff = (sl - slr) / (sls - slr)
                pcap = -1.0 / abs(alpha) * (Seff ** (-1.0 / m) - 1.0) ** (1.0 - m)
            else:
                pcap = -abs(pmax)
            pcap = max(pcap, -abs(pmax))
            pcap *= (1.0 - sl) / 0.001 if sl > 0.999 else 1.0
        return pcap

    @property
    def parameters(self):
        """Return model parameters."""
        return [self._m, self._slr, self._alpha, self._pmax, self._sls]

    @parameters.setter
    def parameters(self, value):
        assert len(value) == 5
        self._m = value[0]
        self._slr = value[1]
        self._alpha = value[2]
        self._pmax = value[3]
        self._sls = value[4]



from ._base import BaseRelativePermeability

__all__ = [
    "Verma",
]


class Verma(BaseRelativePermeability):
    """Verma's function.

    After Verma et al. (1985).

    Parameters
    ----------
    slr : scalar
        Irreducible liquid saturation (RP(1)).
    sls : scalar
        Maximum liquid saturation (RP(2)).
    a : scalar
        A (RP(3)).
    b : scalar
        B (RP(4)).
    c : scalar
        C (RP(5)).

    """

    _id = 8
    _name = "Verma"

    def __init__(self, slr=0.2, sls=0.895, a=1.259, b=-1.7615, c=0.5089):
        self.parameters = [slr, sls, a, b, c]

    def _eval(self, sl, slr, sls, a, b, c):
        """Verma's function."""
        Shat = (sl - slr) / (sls - slr)
        Shat = max(Shat, 0.0)
        Shat = min(Shat, 1.0)
        kl = Shat ** 3
        kg = a + b * Shat + c * Shat ** 2
        kg = max(kg, 0.0)
        kg = min(kg, 1.0)
        return kl, kg

    @property
    def parameters(self):
        """Return model parameters."""
        return [self._slr, self._sls, self._a, self._b, self._c]

    @parameters.setter
    def parameters(self, value):
        assert len(value) == 5
        self._slr = value[0]
        self._sls = value[1]
        self._a = value[2]
        self._b = value[3]
        self._c = value[4]



from ._base import BaseCapillarity

__all__ = [
    "Milly",
]


class Milly(BaseCapillarity):
    """Milly's function.

    After Milly (1982).

    Parameters
    ----------
    slr : scalar
        Irreducible liquid saturation (CP(1)).

    """

    _id = 4
    _name = "Milly"

    def __init__(self, slr):
        assert slr >= 0.0
        self.parameters = [slr]

    def _eval(self, sl, slr):
        sl = max(sl, 1.001 * slr)
        fac = (
            1.0
            if sl - slr >= 0.371
            else 10 ** (2.26 * (0.371 / (sl - slr) - 1.0) ** 0.25 - 2.0)
        )
        return -97.783 * fac

    @property
    def parameters(self):
        """Return model parameters."""
        return [self._slr]

    @parameters.setter
    def parameters(self, value):
        assert len(value) == 1
        self._slr = value[0]



import numpy

from ._base import BaseCapillarity

__all__ = [
    "Pickens",
]


class Pickens(BaseCapillarity):
    """Pickens et al. function.

    After Pickens et al. (1979).

    Parameters
    ----------
    p0 : scalar
        Capillary pressure strength (CP(1)).
    slr : scalar
        Irreducible liquid saturation (CP(2)).
    sl0 : scalar
        Initial liquid saturation (CP(3)).
    x : scalar
        CP(4).

    """

    _id = 2
    _name = "Pickens"

    def __init__(self, p0, slr, sl0, x):
        assert 0.0 < slr < 1.0
        assert sl0 >= 1.0
        assert x != 0.0
        self.parameters = [p0, slr, sl0, x]

    def _eval(self, sl, p0, slr, sl0, x):
        """Pickens et al. function."""
        sl = max(sl, 1.001 * slr)
        sl = 0.999 * sl0 if sl > 0.999 * sl0 else sl

        A = (1.0 + sl / sl0) * (sl0 - slr) / (sl0 + slr)
        B = 1.0 - sl / sl0
        return -p0 * (numpy.log(A / B * (1.0 + (1.0 - B ** 2 / A ** 2) ** 0.5))) ** (
            1.0 / x
        )

    @property
    def parameters(self):
        """Return model parameters."""
        return [self._p0, self._slr, self._sl0, self._x]

    @parameters.setter
    def parameters(self, value):
        assert len(value) == 4
        self._p0 = value[0]
        self._slr = value[1]
        self._sl0 = value[2]
        self._x = value[3]



from __future__ import print_function

__all__ = [
    "export",
]


format_to_ext = {
    "tecplot": ".dat",
    "vtk": ".vtk",
    "vtu": ".vtu",
}


def export(argv=None):
    import os
    import sys
    from .. import read_mesh, read_output
    from ..meshmaker import voxelize, triangulate

    parser = _get_parser()
    args = parser.parse_args(argv)

    # Check that TOUGH output and mesh file exist
    assert os.path.isfile(args.infile), "TOUGH output file '{}' not found.".format(
        args.infile
    )
    if args.mesh:
        assert os.path.isfile(args.mesh), "Pickled mesh file '{}' not found.".format(
            args.mesh
        )

    # Read output file
    print("Reading file '{}' ...".format(args.infile), end="")
    sys.stdout.flush()
    out = read_output(args.infile)
    if args.time_step is not None:
        assert -len(out) <= args.time_step < len(out), "Inconsistent time step value."
        out = out[args.time_step]
    else:
        out = out[-1]
    print(" Done!")

    # Triangulate or voxelize if no mesh
    if not args.mesh:
        print("Mesh file not specified, inferring dimensionality ...", end="")
        sys.stdout.flush()
        points = _get_points(out)
        ndim = 1 if points.ndim == 1 else points.shape[1]
        print(" Done!")

        if ndim > 1:
            print(
                "Mesh is {}D, performing point triangulation ...".format(ndim), end=""
            )
            sys.stdout.flush()

            mesh = triangulate(points)
            for label, data in out.data.items():
                if label not in {"X", "Y", "Z"}:
                    mesh.add_point_data(label, data)

        else:
            print("Mesh is 1D, voxelizing points ...", end="")
            sys.stdout.flush()

            mesh = voxelize(points)
            for label, data in out.data.items():
                if label not in {"X", "Y", "Z"}:
                    mesh.add_cell_data(label, data)

    else:
        print("Reading mesh file '{}' ...".format(args.mesh), end="")
        sys.stdout.flush()
        try:
            mesh = read_mesh(args.mesh, file_format="pickle")
        except:
            raise ValueError("Cannot unpickle mesh file {}.".format(args.mesh))
        mesh.read_output(out)
    print(" Done!")

    # Output file name
    if args.output_file:
        filename = args.output_file
    else:
        head, _ = os.path.splitext(args.infile)
        filename = "{}{}".format(head, format_to_ext[args.file_format])

    # Write output file
    print("Writing output file '{}' ...".format(filename), end="")
    sys.stdout.flush()
    mesh.cell_data.pop("material")
    mesh.write(filename, file_format=args.file_format)
    print(" Done!")


def _get_parser():
    import argparse

    # Initialize parser
    parser = argparse.ArgumentParser(
        description=("Export TOUGH simulation results to a file for visualization."),
        formatter_class=argparse.RawTextHelpFormatter,
    )

    # Input file
    parser.add_argument(
        "infile", type=str, help="TOUGH output file",
    )

    # Mesh file
    parser.add_argument(
        "--mesh", "-m", type=str, default=None, help="Pickled toughio.Mesh",
    )

    # Time step
    parser.add_argument(
        "--time-step", "-t", type=int, default=None, help="Time step to export",
    )

    # Output file
    parser.add_argument(
        "--output-file", "-o", type=str, default=None, help="Exported file",
    )

    # File format
    parser.add_argument(
        "--file-format",
        "-f",
        type=str,
        choices=("tecplot", "vtk", "vtu"),
        default="vtk",
        help="Exported file format",
    )

    return parser


def _get_points(data):
    import numpy

    # Make sure that point coordinates exist
    assert "X" in data.data.keys(), "Data 'X' not found."
    assert "Y" in data.data.keys(), "Data 'Y' not found."
    assert "Z" in data.data.keys(), "Data 'Z' not found."

    X = numpy.array(data.data["X"])
    Y = numpy.array(data.data["Y"])
    Z = numpy.array(data.data["Z"])

    # Assert dimension of the problem
    nx = len(numpy.unique(X))
    ny = len(numpy.unique(Y))
    nz = len(numpy.unique(Z))

    # Reconstruct points cloud
    points = (
        X
        if ny == 1 and nz == 1
        else Y
        if nx == 1 and nz == 1
        else Z
        if nx == 1 and ny == 1
        else numpy.column_stack((Y, Z))
        if nx == 1
        else numpy.column_stack((X, Z))
        if ny == 1
        else numpy.column_stack((X, Y))
        if nz == 1
        else numpy.column_stack((X, Y, Z))
    )

    return points
